from pyod.models.auto_encoder import AutoEncoder


class AutoEncoderAlgorithm:
    algorithm = AutoEncoder
    name = "AutoEncoder"
    import_code = "from pyod.models.auto_encoder import AutoEncoder"

    def __init__(self):
        pass

    @classmethod
    def get_arg(cls, **args):
        dataframe = args['dataframe']
        return {
            'hidden_neurons': [len(dataframe.columns), 32, 32, len(dataframe.columns)]
        }

    @classmethod
    def get_args_code(cls, **args):
        dataframe = args['dataframe']
        return 'hidden_neurons = [{0}, 32, 32, {0}]'.format(len(dataframe.columns)),



import os
import pickle
import pandas as pd
from sklearn.tree import DecisionTreeClassifier
from sklearn.preprocessing import LabelEncoder


class SuggestionManager:

    def __init__(self, save_path, commands, force_training=False):
        self.path = save_path
        self.log = []

        self.possible_commands = ['None']
        for key in commands:
            self.possible_commands.append(key)
        self.y_encoder = LabelEncoder()
        self.y_encoder.fit(self.possible_commands)
        self.config = {
            "counter": 0,
            "limit": 1000,
            "data": pd.DataFrame(columns=[
                "cmd1","cmd2","cmd3","cmd4","cmd5",
                'speaking', 'algorithm', 'preprocessing',
                'visualization', 'pipe', 'metric',
                'code', 'other', 'clf'
            ])
        }

        if os.path.exists(self.path) and force_training == False:
            with open(self.path, 'rb') as file:
                self.config = pickle.load(file)
        else:
            with open(self.path, 'wb') as file:
                pickle.dump(self.config, file)

        self.model = DecisionTreeClassifier(random_state=0)
        if len(self.config['data']) > 0:
            self.train_model()

    def add(self, cmd , user_preferences):

        # print(self.predict(user_preferences))

        # check for high values in the user preferences
        found = False
        for tag, value in user_preferences.items():
            if value > 0.5:
                found = True
        # check if we found a high value and there is enough  train data
        # if there is not enough train data, the line will be added as well
        if not found and self.config["counter"] > self.config["limit"]:
            # add the current command for further analysis
            self.log.append(cmd.tag)
            return

        # get the last 5 commands
        current_line = ['None', 'None', 'None', 'None', 'None']
        n = len(self.log)
        for i in range(0, min(n, 5)):
            log_index = (n-i)-1
            line_index = (5-i)-1
            current_line[line_index] = self.log[log_index]

        current_line.append(round(user_preferences["speaking"],3))
        current_line.append(round(user_preferences["algorithm"],3))
        current_line.append(round(user_preferences["preprocessing"],3))
        current_line.append(round(user_preferences["visualization"],3))
        current_line.append(round(user_preferences["pipe"],3))
        current_line.append(round(user_preferences["metric"],3))
        current_line.append(round(user_preferences["code"],3))
        current_line.append(round(user_preferences["other"],3))

        # add the classification at the end of the line
        current_line.append(cmd.tag)

        # add the current command for further analysis
        self.log.append(cmd.tag)

        # append a line to the dataset or override an old line
        data_index = self.config["counter"] % self.config["limit"]
        self.config["data"].loc[data_index] = current_line
        self.config["counter"] = self.config["counter"] +1

        # train a new suggestion model every 100 new lines
        if self.config["counter"] % 100 == 0:
            self.train_model()

        # save the suggestion file
        with open(self.path, 'wb') as file:
            pickle.dump(self.config, file)

    def train_model(self):
        data = self.config["data"]

        X = data.iloc[:,:-1]
        y = data.iloc[:,-1]

        y = self.y_encoder.transform(y)

        X['cmd1'] = self.y_encoder.transform(X['cmd1'])
        X['cmd2'] = self.y_encoder.transform(X['cmd2'])
        X['cmd3'] = self.y_encoder.transform(X['cmd3'])
        X['cmd4'] = self.y_encoder.transform(X['cmd4'])
        X['cmd5'] = self.y_encoder.transform(X['cmd5'])

        self.model.fit(X, y)

    def predict(self, user_preferences):

        if len(self.config['data']) < 10:
            return self.possible_commands[0]

        # get the last 5 commands
        current_line = ['None', 'None', 'None', 'None', 'None']
        n = len(self.log)
        for i in range(0, min(n, 5)):
            log_index = (n - i) - 1
            line_index = (5 - i) - 1
            current_line[line_index] = self.log[log_index]

        current_line = list(self.y_encoder.transform(current_line))

        current_line.append(round(user_preferences["speaking"], 3))
        current_line.append(round(user_preferences["algorithm"], 3))
        current_line.append(round(user_preferences["preprocessing"], 3))
        current_line.append(round(user_preferences["visualization"], 3))
        current_line.append(round(user_preferences["pipe"], 3))
        current_line.append(round(user_preferences["metric"], 3))
        current_line.append(round(user_preferences["code"], 3))
        current_line.append(round(user_preferences["other"], 3))

        y_pred = self.model.predict([current_line])
        command = self.y_encoder.inverse_transform(y_pred)
        return command



import numpy as np

from .utils.BotConfig import BotConfig
from .commands.base.speech import Speech


class DSBot:

    def __init__(self, **kwargs):
        self.initial_config = kwargs
        self.config = BotConfig(**self.initial_config)
        self.speech_manager = Speech(self.config)
        self.bot = self.config.bot
        self.corpus = self.config.corpus
        self.commands = self.config.commands

    def start(self):
        print("Começe a falar com o bot!")

        while True:
            # print("Escutando ...")
            text = self.config.input()

            # case input control did not undestand
            if not text:
                print("Bot: Nao entendi. Pode tentar novamente.")
                continue

            # print("You: {0}".format(text))

            response = self.speech_manager.next(text)

            tag = response["tag"]
            message = response["message"]

            # print("Bot: {0}".format(response))
            print("Bot: {0}".format(message))

            # exit the while loop
            if tag == 'goodbye':
                self.speech_manager.print()
                # break

        print("Done!")

    def test(self):
        print("Running initial test set ...")

        testing_set = {
            "asfasdf": "unknown",
            "lnvr": "unknown",
            "ewetrt": "unknown",
            "asdfww": "unknown",
            "Vamos carregar o dataset outliers_sample_data em data/outliers.csv": "load_dataset",
            "Selecciona a coluna target e guarda em outliers_target": "column_extract",
            "Remove a coluna target do dataset outliers_sample_data": "column_deletion",
            "Roda o algoritmo loda e guarda em autoencoder_model": "algorithm",
            "Vamos a predizer usando o modelo autoencoder_model e guarda em model_predictions": "predict",
            "Desenha um grafico de pontos com as cores iguais a model_predictions": "scatter plot",
            "Salva o codigo em data/sample.py": "generate_code",
            "Formatar os dados da coluna x1 do dataset cleaning_sample_data para letra minuscula": "lower_upper_case",
            "ordene os dados da coluna x1 do dataset cleaning_sample_data de maneira crescente": "sort_column"
        }

        for key, command in self.commands.items():
            for pattern in command.patterns:
                testing_set[pattern] = command.tag

        total_tests = len(testing_set)

        correct_threshold_5 = 0
        wrong_cases_5 = []

        correct_threshold_6 = 0
        wrong_cases_6 = []

        correct_threshold_7 = 0
        wrong_cases_7 = []

        for pattern, tag in testing_set.items():
            wrds = self.corpus.encode(pattern).reshape(1, -1)
            results = self.bot.answer(wrds)
            max_result = np.max(results)

            results_index = np.argmax(results)
            predicted_tag = self.corpus.labels[results_index]

            if max_result > .5:
                if predicted_tag == tag:
                    correct_threshold_5 += 1
                else:
                    wrong_cases_5.append(pattern)
            else:
                if 'unknown' == tag:
                    correct_threshold_5 += 1
                else:
                    wrong_cases_5.append(pattern)

            if max_result > .6:
                if predicted_tag == tag:
                    correct_threshold_6 += 1
                else:
                    wrong_cases_6.append(pattern)
            else:
                if 'unknown' == tag:
                    correct_threshold_6 += 1
                else:
                    wrong_cases_6.append(pattern)

            if max_result > .7:
                if predicted_tag == tag:
                    correct_threshold_7 += 1
                else:
                    wrong_cases_7.append(pattern)
            else:
                if 'unknown' == tag:
                    correct_threshold_7 += 1
                else:
                    wrong_cases_7.append(pattern)

        print("threshold 0.5 : {0}/{1}".format(correct_threshold_5, total_tests))
        print("threshold 0.6 : {0}/{1}".format(correct_threshold_6, total_tests))
        print("threshold 0.7 : {0}/{1}".format(correct_threshold_7, total_tests))

        print("Failed 0.5: {0}".format(wrong_cases_5))
        print("Failed 0.6: {0}".format(wrong_cases_6))
        print("Failed 0.7: {0}".format(wrong_cases_7))

    def register(self, cmd):
        # add command to the config
        if 'commands' not in self.initial_config:
            self.initial_config['commands'] = self.config.commands
        self.initial_config['commands'][cmd.tag] = cmd

    def train(self):
        old_force_training = self.config.force_training
        self.initial_config['force_training']= True

        self.config = BotConfig(**self.initial_config)
        self.speech_manager = Speech(self.config)
        self.bot = self.config.bot
        self.corpus = self.config.corpus
        self.commands = self.config.commands

        self.config.force_training = old_force_training




from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class CellEmptyCommand(CommandWithArgs):

    tag = "empty_cell"
    patterns = [
        "substitua os espaços brancos",
        "troque os espaços vazios",
        "troque os espaços em branco",
        "troque os valores null",
        "substitua os valores nan",
    ]

    def __init__(self, parent, task_manager):
        super(CellEmptyCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso substitua os espaços vazios da coluna {x} pela {y}."]
        self.user_config_tag = 'preprocessing'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.column_name = Argument({
            'parent': self,
            'name': 'column_name',
            'trigger': 'coluna',
            'position': 1
        })

        self.cell_value = Argument({
            "parent": self,
            "name": 'cell_value',
            "required": True,
            "position": 1,
            "patterns": [
                'pela (mediana|media)',
                'por a (mediana|media)'
            ]
        })

        self.children = [
            self.dataset_name,
            self.column_name,
            self.cell_value
        ]

    def run(self, context):
        import numpy as np

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value
        cell_value = self.cell_value.value

        x = context[dataset_name][column_name]
        column_index=context[dataset_name].columns.get_loc(column_name)

        sum = 0
        total = len(x)
        wrong_lines = []
        for index, value in x.items():
            try:
                current= float(value)
                if not np.isnan(current):
                    # print("{0}: {1}".format(index,current))
                    sum = sum + current
                else:
                    wrong_lines.append(index)
            except:
                wrong_lines.append(index)

        replace = 0
        if cell_value == "media":
            replace = sum/total

        # replace wrong values.
        context[dataset_name].iloc[wrong_lines, column_index] = replace


    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Select column from dataset")

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value
        cell_value = self.cell_value.value

        # code_generator.write("{0} = {1}['{2}']".format(context_variable, dataset_name, column_name))




import numpy as np
import os
import pickle

from .datascience_glossary import DataScienceGlossary


class ExpertiseAnalyzer():
    def __init__(self):
        self.path = 'data/user.settings.pickle'
        self.user_config = {
            'text_expertise': 0.5,
            'commands':{
                'speaking': 0.125,
                'algorithm': 0.125,
                'preprocessing': 0.125,
                'visualization': 0.125,
                'pipe': 0.125,
                'metric': 0.125,
                'code': 0.125,
                'other': 0.125
            }
        }
        if not os.path.exists(self.path):
            with open(self.path, 'wb') as file:
                pickle.dump(self.user_config, file)
        else:
            with open(self.path, 'rb') as file:
                self.user_config = pickle.load(file)

        self.analyzers = [
            DataScienceGlossary()
        ]

    def analyze(self, sentence):
        for analyzer in self.analyzers:
            analyzer.analyze(sentence, self.user_config)

        # update user settings file
        with open(self.path, 'wb') as file:
            pickle.dump(self.user_config, file)





from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument

class ColumnExtractCommand(CommandWithArgs):

    tag = "column_extract"
    patterns = [
        "Guarda a coluna",
        "Seleciona uma coluna",
        "Sava a columna",
    ]

    def __init__(self, parent, task_manager):
        super(ColumnExtractCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso extrair a coluna {x} e guardar em {y}."]
        self.user_config_tag = 'preprocessing'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.column_name = Argument({
            'parent': self,
            'name': 'column_name',
            'trigger': 'coluna',
            'position': 1
        })

        self.context_variable = Argument({
            'parent': self,
            'name': 'variable_name',
            'trigger': 'guarda em',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.column_name,
            self.context_variable
        ]

    def run(self, context):

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value
        context_variable = self.context_variable.value

        x = context[dataset_name][column_name]
        context[context_variable] = x

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Select column from dataset")

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value
        context_variable = self.context_variable.value

        code_generator.write("{0} = {1}['{2}']".format(context_variable, dataset_name, column_name))




import tensorflow as tf
# print(tf.__version__)


class Dense128NetBot:

    def __init__(self, model_path, corpus="", verbose=True, force_training=False):
        self.verbose = verbose
        self.model_path = model_path
        self.corpus = corpus

        if force_training:
            self.load_model()
        else:
            try:
                if self.verbose:
                    print("Procurando um modelo treinado ...")
                self.model = tf.keras.models.load_model(model_path)
            except:
                if self.verbose:
                    print("Nao foi possivel encontrar um modelo treinado. Treinando um modelo novo ...")
                    print("Amostras de treinamento: {0}".format(len(self.corpus.training[0])))
                self.load_model()

    def load_model(self):

        self.model = tf.keras.models.Sequential([
            tf.keras.layers.InputLayer(input_shape=(len(self.corpus.training[0]),)),
            tf.keras.layers.Dropout(0.5),
            tf.keras.layers.Dense(128),
            tf.keras.layers.Dropout(0.3),
            tf.keras.layers.Dense(64),
            tf.keras.layers.Dropout(0.5),
            tf.keras.layers.Dense(32),
            # tf.keras.layers.Dropout(0.5),
            tf.keras.layers.Dense(len(self.corpus.output[0]), activation='softmax')
        ])

        self.model.compile(
            optimizer=tf.keras.optimizers.SGD(),
            loss=tf.keras.losses.CategoricalCrossentropy(from_logits=False),
            metrics=['accuracy']
        )

        self.model.fit(self.corpus.training, self.corpus.output, epochs=1000, batch_size=16, verbose=0)

        self.model.save(self.model_path)

    def answer(self, user_speech):
        return self.model.predict(user_speech)



from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class PredictCommand(CommandWithArgs):

    tag = "predict"
    patterns = [
        "queria predizer",
        "prediz usando o modelo",
        "usa o modelo para predizer"
    ]

    def __init__(self, parent, task_manager):
        super(PredictCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso usar o modelo {0} para predizer o dataset {1} e guardar o resultao em {2}."]
        self.user_config_tag = 'algorithm'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.model_name = Argument({
            'parent': self,
            'name': 'model_name',
            'trigger': 'modelo',
            'position': 1
        })

        self.context_variable = Argument({
            'parent': self,
            'name': 'variable_name',
            'trigger': 'guarda em',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.model_name,
            self.context_variable
        ]

    def verify_params(self, context):
        model_name = self.model_name.value
        model = context[model_name]
        if not hasattr(model, "predict"):
            self.model_name.complete=False
            return [{
                'node': self.model_name,
                'tag': 'model_name',
                'message': '{0} nao tem um metodo predict e nao pode ser usado para predizer'.format(model_name)
            }]
        return []

    def run(self, context):
        dataset_name = self.dataset_name.value
        model_name = self.model_name.value
        context_variable = self.context_variable.value

        dataframe = context[dataset_name]
        model = context[model_name]

        # result = model.predict(dataframe.to_numpy())
        result = model.predict(dataframe)

        context[context_variable] = result

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# call algorithm predict method")

        dataset_name = self.dataset_name.value
        model_name = self.model_name.value
        context_variable = self.context_variable.value

        code_generator.write("{0} = {1}.predict({2})".format(context_variable, model_name, dataset_name))




from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class HistogramCommand(CommandWithArgs):

    tag = "histogram"
    patterns = [
        "Pode fazer um histograma",
        "Eu quero um histograma",
        "Desenha um histograma",
        "faz um histograma",
        "histograma",
    ]

    def __init__(self, parent, task_manager):
        super(HistogramCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso fazer um histograma com {x}."]
        self.user_config_tag = 'visualization'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.x_axis = Argument({
            'parent': self,
            'name': 'x',
            'trigger': 'eixo x igual a',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.x_axis
        ]

    def run(self, context):
        import matplotlib.pyplot as plt
        plt.clf()

        dataset_name = self.dataset_name.value
        x_arg = self.x_axis.value
        x = context[dataset_name][x_arg]

        plt.hist(x, bins=10)
        # plt.show()
        plt.savefig("plot_result.png")

        image = open("plot_result.png", 'rb')
        context["plot_result"] = image

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Histogram ")
        code_generator.write("import matplotlib.pyplot as plt")

        dataset_name = self.dataset_name.value
        x_arg = self.x_axis.value
        x_arg_name = self.x_axis.name

        code_generator.write("{0} = {1}['{2}']".format(x_arg_name, dataset_name, x_arg))
        code_generator.write("plt.hist({0}, bins=10)".format(x_arg_name))
        code_generator.write("plt.show()")




from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class SortColumnCommand(CommandWithArgs):

    tag = "sort_column"
    patterns = [
        "Ordene os valores decrescente",
        "Ordene os valores crescente",
        "Ordene os dados crescente",
        "Ordene os dados decrescente",
    ]

    def __init__(self, parent, task_manager):
        super(SortColumnCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso ordenar pela coluna {x}"]
        self.user_config_tag = 'preprocessing'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.column_name = Argument({
            'parent': self,
            'name': 'column_name',
            'trigger': 'coluna',
            'position': 1
        })

        self.cell_value = Argument({
            "parent": self,
            "name": 'order',
            "required": True,
            "position": 1,
            "patterns": [
                '(decrescente|crescente)',
                '(de maior a menor|de menor a maior)',
            ]
        })

        self.children = [
            self.dataset_name,
            self.column_name,
            self.cell_value
        ]

    def run(self, context):

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value
        cell_value = self.cell_value

        df = context[dataset_name]

        ascending = True
        if cell_value in ["decrescente", "de maior a menor"]:
            ascending= False

        try:
            df.sort_values(by=[column_name], ascending=ascending, inplace=True)
        except Exception as e:
            print(e)

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Select column from dataset")

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        # code_generator.write("{0} = {1}['{2}']".format(context_variable, dataset_name, column_name))




from pyod.models.cblof import CBLOF


class CblofAlgorithm:
    algorithm = CBLOF
    name = "CBLOF"
    import_code = "from pyod.models.cblof import CBLOF"

    def __init__(self):
        pass

    @classmethod
    def get_arg(cls, **args):
        return {}

    @classmethod
    def get_args_code(cls, **args):
        return ''




class Dialog:
    def __init__(self, dialog = [], verbose=True):
        self.dialog = dialog
        self.index = -1
        self.verbose = verbose

    def input(self):
        self.index += 1
        try:
            return self.dialog[self.index]
        except Exception as e:
            return input()

    def __call__(self, *args, **kwargs):
        return self.input()



from pyod.models.knn import KNN


class KnnAlgorithm:
    algorithm = KNN
    name = "KNN"
    import_code = "from pyod.models.knn import KNN"

    def __init__(self):
        pass

    @classmethod
    def get_arg(cls, **args):
        return {}

    @classmethod
    def get_args_code(cls, **args):
        return ''



from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class AccuracyCommand(CommandWithArgs):

    tag = "metrics_accuracy"
    patterns = [
        "calcula a acuracia",
        "accuracia",
        "qual foi a acuracia",
    ]

    def __init__(self, parent, task_manager):
        super(AccuracyCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso calcular a acuracia e guardar em {0}."]
        self.user_config_tag = 'metric'

        self.complete = False

        self.y_true_arg = Argument({
            'parent': self,
            'name': 'y_true',
            'trigger': 'y_true',
            'position': 1
        })

        self.y_pred_arg = Argument({
            'parent': self,
            'name': 'y_pred',
            'trigger': 'y_pred',
            'position': 1
        })

        self.context_variable = Argument({
            'parent': self,
            'name': 'variable_name',
            'trigger': 'guarda em',
            'position': 1
        })

        self.children = [
            self.y_true_arg,
            self.y_pred_arg,
            self.context_variable
        ]

    def run(self, context):

        y_true_arg = self.y_true_arg.value
        y_pred_arg = self.y_pred_arg.value
        context_variable = self.context_variable.value

        y_true = context[y_true_arg]
        y_pred = context[y_pred_arg]

        from sklearn.metrics import accuracy_score
        result = accuracy_score(y_true, y_pred)
        context[context_variable] = result

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Accuracy Score")

        y_true_arg = self.y_true_arg.value
        y_pred_arg = self.y_pred_arg.value
        context_variable = self.context_variable.value

        code_generator.write("from sklearn.metrics import accuracy_score")
        code_generator.write("{0} = accuracy_score({1}, {2})".format(context_variable, y_true_arg, y_pred_arg))




from pyod.models.lmdd import LMDD


class LmddAlgorithm:
    algorithm = LMDD
    name = "LMDD"
    import_code = "from pyod.models.lmdd import LMDD"

    def __init__(self):
        pass

    @classmethod
    def get_arg(cls, **args):
        return {}

    @classmethod
    def get_args_code(cls, **args):
        return ''



import random
from sklearn import datasets

from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument
from ..base.predict import PredictCommand

from ..preprocess.extract_column import ColumnExtractCommand
from ..preprocess.remove_column import ColumnRemoveCommand

from ..metrics.accuracy import AccuracyCommand
from ..algoritms.pyod import *

from ..visualization.plot_scatter import ScatterPlotCommand


class OutliersPipeCommand(CommandWithArgs):

    tag = "outliers_pipe"
    patterns = [
        "analise de pontos extremos",
        "quais são os pontos extremos",
        "pontos extremos"
    ]

    def __init__(self, parent, task_manager):
        super(OutliersPipeCommand, self).__init__(parent, task_manager)
        self.responses = ["Fazendo uma analise de pontos extremos no dataset {0}."]
        self.user_config_tag = 'pipe'

        self.algorithms = {
            "abod": AbodAlgorithm,
            "pyod_autoencoder": AutoEncoderAlgorithm,
            "pyod_cblof": CblofAlgorithm,
            "pyod_hbos": HbosAlgorithm,
            "pyod_iforest": IForestAlgorithm,
            "pyod_knn": KnnAlgorithm,
            "pyod_lmdd": LmddAlgorithm,
            "pyod_loda": LodaAlgorithm,
        }

        self.complete = False

        self.dataset_name_arg = Argument({
            "parent": self,
            "name": 'dataset_name',
            "position": 1,
            "trigger": "dataset"
        })

        self.target_arg = Argument({
            "parent": self,
            "name": 'outline_clf',
            "position": 1,
            # Never extract from input text
            "trigger": "outlier classification column"
        })

        self.children = [
            self.dataset_name_arg,
            self.target_arg
        ]

    def run(self, context):
        dataset_name = self.dataset_name_arg.value
        dataframe = context[dataset_name]
        target = self.target_arg.value

        column_extraction = ColumnExtractCommand(parent=self, task_manager=self.task_manager)
        column_extraction.dataset_name = self.dataset_name_arg
        column_extraction.column_name = self.target_arg
        column_extraction.context_variable.value = "outliers_pipe_target"
        column_extraction.context_variable.complete = True
        column_extraction.complete = True

        column_extraction.run(context)
        # self.task_manager.add_task({
        #     'fn': column_extraction.run,
        #     'args': {
        #         'context': context,
        #     }
        # })

        self.children.append(column_extraction)

        column_deletion = ColumnRemoveCommand(parent=self, task_manager=self.task_manager)
        column_deletion.dataset_name = self.dataset_name_arg
        column_deletion.column_name = self.target_arg
        column_deletion.complete = True

        column_deletion.run(context)
        # self.task_manager.add_task({
        #     'fn': column_deletion.run,
        #     'args': {
        #         'context': context,
        #     }
        # })

        self.children.append(column_deletion)


        algorithms_accuracy = []
        best_acc = -1
        best_alg = ""
        for current_algorithm_name, current_algorithm in self.algorithms.items():
            algorithm = current_algorithm.algorithm
            algorithm = algorithm(current_algorithm.get_arg(dataframe=dataframe))
            algorithm.fit(dataframe)
            context["{0}_outliers_pipe_model".format(current_algorithm_name)] = algorithm

            predict = PredictCommand(parent=self, task_manager=self.task_manager)
            predict.dataset_name = self.dataset_name_arg
            predict.model_name.value = "{0}_outliers_pipe_model".format(current_algorithm_name)
            predict.model_name.complete = True
            predict.context_variable.value = "{0}_outliers_pipe_model_prediction".format(current_algorithm_name)
            predict.context_variable.complete = True
            predict.complete = True

            predict.run(context)
            # self.task_manager.add_task({
            #     'fn': predict.run,
            #     'args': {
            #         'context': context,
            #     }
            # })

            self.children.append(predict)

            accuracy = AccuracyCommand(parent=self, task_manager=self.task_manager)
            accuracy.y_true_arg.value = "outliers_pipe_target"
            accuracy.y_true_arg.complete = True
            accuracy.y_pred_arg.value = "{0}_outliers_pipe_model_prediction".format(current_algorithm_name)
            accuracy.y_pred_arg.complete = True
            accuracy.context_variable.value = "{0}_outliers_pipe_model_accuracy".format(current_algorithm_name)
            accuracy.context_variable.complete = True
            accuracy.complete = True

            accuracy.run(context)
            # self.task_manager.add_task({
            #     'fn': accuracy.run,
            #     'args': {
            #         'context': context,
            #     }
            # })

            self.children.append(accuracy)

            acc_value =context[accuracy.context_variable.value]
            algorithms_accuracy.append({
                "name":current_algorithm_name,
                "accuracy": acc_value
            })

            # update best result
            if(acc_value > best_acc):
                self.report("{0} obteve um resultado de {1} MELHOR do que o {2} do algoritmo {3}".format(current_algorithm_name,acc_value, best_acc, best_alg ))
                best_acc = acc_value
                best_alg = current_algorithm_name
            else:
                self.report("{0} deu {1} mais NAO melhorou o {2} do algoritmo {3}".format(current_algorithm_name,acc_value, best_acc, best_alg ))


            scatter_plot = ScatterPlotCommand(parent=self, task_manager=self.task_manager)
            scatter_plot.title = "{0} Algorithm".format(current_algorithm_name)
            scatter_plot.dataset_name = self.dataset_name_arg
            scatter_plot.colors.value = "{0}_outliers_pipe_model_prediction".format(current_algorithm_name)
            scatter_plot.colors.complete = True
            scatter_plot.x_axis.value = 'x1'
            scatter_plot.x_axis.complete = True
            scatter_plot.y_axis.value = 'x2'
            scatter_plot.y_axis.complete = True
            scatter_plot.complete = True

            scatter_plot.run(context)
            # self.task_manager.add_task({
            #     'fn': scatter_plot.run,
            #     'args': {
            #         'context': context,
            #     }
            # })

            self.children.append(scatter_plot)

        # Finally print the results
        # algorithms_accuracy = sorted(algorithms_accuracy, key=lambda k: k['accuracy'], reverse=True)
        # self.print_metrics(algorithms_accuracy)

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Begin: Outlines Analysis Pipeline")
        for child in self.children:
            child.generate_code(code_generator, context)
        code_generator.write("")
        code_generator.write("# End: Outlines Analysis Pipeline")

    def print_metrics(self, metrics):
        print("\tName\t|\tAccuracy")
        print("\t--------------------")
        for alg in metrics:
            print("\t{0}\t|\t{1}".format(alg['name'], alg['accuracy']))


from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class StatisticalCorrelationCommand(CommandWithArgs):

    tag = "statistical_correlation"
    patterns = [
        "pode fazer uma correlação estatistica",
        "correlação estatistica",
        "faz uma correlação estatistica"
    ]

    def __init__(self, parent, task_manager):
        super(StatisticalCorrelationCommand, self).__init__(parent, task_manager)
        self.responses = ["Eu posso fazer uma correlação estatistica emtre {x} e {y}."]
        self.user_config_tag = 'preprocessing'
        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.x_axis = Argument({
            'parent': self,
            'name': 'x',
            'trigger': 'eixo x igual a',
            'position': 1
        })

        self.y_axis = Argument({
            'parent': self,
            'name': 'y',
            'trigger': 'eixo y igual a',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.x_axis,
            self.y_axis,
        ]

    def run(self, context):
        from scipy import stats

        dataset_name = self.dataset_name.value

        x_arg = self.x_axis.value
        x = context[dataset_name][x_arg]

        y_arg = self.y_axis.value
        y = context[dataset_name][y_arg]

        result = stats.pearsonr(x, y)
        print(result)

    def generate_code(self, code_generator, context):

        code_generator.write("")
        code_generator.write("# Pearson Statistical Correlation ")
        code_generator.write("from scipy import stats")

        x_arg = self.x_axis.value
        x_arg_name = self.x_axis.name

        y_arg = self.y_axis.value
        y_arg_name = self.y_axis.name

        dataset_name_arg = self.dataset_name.value

        code_generator.write("{0} = {1}['{2}']".format(x_arg_name, dataset_name_arg, x_arg))
        code_generator.write("{0} = {1}['{2}']".format(y_arg_name, dataset_name_arg, y_arg))
        code_generator.write("result = stats.pearsonr({0}, {1})".format(x_arg_name, y_arg_name))
        code_generator.write("print(result)")


from .command_with_args import CommandWithArgs
from ..base.argument import Argument


class GenerateScriptCommand(CommandWithArgs):

    tag = "generate_code"
    patterns = [
        "Gera um script",
        "Salva um arquivo com o codigo fonte",
        "Salva o meu script",
        "Gera o codigo"
    ]

    def __init__(self, parent, task_manager):
        super(GenerateScriptCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso gerar o script em  {script_path}."]
        self.user_config_tag = 'code'
        self.complete = False

        self.script_path_arg = Argument({
            "parent": self,
            "name": 'script_path',
            "required": False,
            "position": 1,
            "patterns": [
                'em ([a-zA-Z0-9_.\\\/]+)',
                'no arquivo ([a-zA-Z0-9_.\\\/]+)',
                'script_path\s*=\s*([a-zA-Z0-9_.\\\/]+)'
            ]
        })

        self.children = [
            self.script_path_arg
        ]

    def run(self, context):
        script_path_arg = self.script_path_arg.value
        print("Generating script in {0}".format(script_path_arg))

        root = self
        while hasattr(root, 'parent'):
            root = root.parent

        code_generator = CodeGeneratorBackend()
        code_generator.write("if __name__ == '__main__':")
        code_generator.indent()
        # code_generator.write("print('Hello Code Generator!')")
        root.generate_code(code_generator, context)
        code = code_generator.end()

        file = open(script_path_arg, 'w')
        file.write(code)
        file.close()


class CodeGeneratorBackend:

    def __init__(self, tab="\t"):
        self.code = []
        self.tab = tab
        self.level = 0

    def end(self):
        return "".join(self.code)

    def write(self, string):
        self.code.append(self.tab * self.level + string + '\n')

    def indent(self):
        self.level = self.level + 1

    def dedent(self):
        if self.level == 0:
            raise SyntaxError("internal error in code generator")
        self.level = self.level - 1


import random
from sklearn import datasets

from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class LoadDatasetCommand(CommandWithArgs):

    tag = "load_dataset"
    patterns = [
        "Quero carregar um dataset",
        "Quero carregar o dataset",
        "utilizar o dataset"
    ]

    def __init__(self, parent, task_manager):
        super(LoadDatasetCommand, self).__init__(parent, task_manager)
        self.responses = ["Dataset carregado."]
        self.user_config_tag = 'preprocessing'

        self.default_datasets={
            'iris': (datasets.load_iris, "datasets.load_iris()"),
            'digits':(datasets.load_digits, "datasets.load_digits()"),
            'wine':(datasets.load_wine, "datasets.load_wine()"),
            # 'airline':(sktime.datasets.load_airline, "datasets.load_airline()"),
            # 'arrow':(sktime.datasets.load_arrow_head, "datasets.load_arrow_head(return_X_y=True)"),
        }

        self.complete = False

        self.dataset_name_arg = Argument({
            "parent": self,
            "name": 'dataset_name',
            "position": 1,
            "patterns": [
                'dataset ([a-zA-Z0-9_]+)',
                'dataset_name\s*=\s*([a-zA-Z0-9_]+)'
            ]
        })

        self.dataset_path_arg = Argument({
            "parent": self,
            "name": 'dataset_path',
            "required": False,
            "position": 1,
            "patterns": [
                'em ([a-zA-Z0-9_.\\\/]+)',
                'no arquivo ([a-zA-Z0-9_.\\\/]+)',
                'dataset_path\s*=\s*([a-zA-Z0-9_.\\\/]+)'
            ]
        })

        self.children = [
            self.dataset_name_arg,
            self.dataset_path_arg
        ]

    def forward(self, text, context):
        if not self.dataset_name_arg.complete:
            return self.dataset_name_arg.forward(text, context)

        dataset_name = self.dataset_name_arg.value

        if dataset_name not in self.default_datasets:
            if not self.dataset_path_arg.complete:
                return self.dataset_path_arg.forward(text, context)

        self.complete = True

        # self.run(context)
        self.task_manager.add_task({
            'fn': self.run,
            'args': {
                'context': context,
            }
        })

        response = random.choice(self.responses)
        return self.parent.backward(self.tag, response, context)

    def run(self, context):
        dataset_name = self.dataset_name_arg.value

        if dataset_name.lower() in self.default_datasets:
            dataframe = self.load_defaults(dataset_name)
            # print(dataframe)
            context[dataset_name] = dataframe
            context[self.dataset_name_arg.name] = dataset_name

        else:
            import pandas as pd
            dataset_path = self.dataset_path_arg.value
            dataframe = pd.read_csv(dataset_path)
            context[dataset_name] = dataframe
            context["last_used_dataset"] = dataframe

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Load Dataset")
        code_generator.write("from sklearn import datasets")
        code_generator.write("import pandas as pd")
        code_generator.write("import numpy as np")

        dataset_name = self.dataset_name_arg.value
        if dataset_name in self.default_datasets:
            _, load_fn_code = self.default_datasets[dataset_name]
            code_generator.write("dataset = {0}".format(load_fn_code))
            code_generator.write("{0} = pd.DataFrame(".format(dataset_name))
            code_generator.indent()
            code_generator.write("data=np.c_[dataset['data'], dataset['target']],")
            code_generator.write(
                "columns= [column_name.replace(' ','') for column_name in dataset['feature_names']] + ['target']")
            code_generator.dedent()
            code_generator.write(")")
        else:
            dataset_path = self.dataset_path_arg.value
            code_generator.write("{0} = pd.read_csv('{1}')".format(dataset_name,dataset_path))

    def load_defaults(self, name):

        import pandas as pd
        import numpy as np

        load_fn , _ = self.default_datasets[name]

        dataset = load_fn()
        if type(dataset) is pd.DataFrame:
            return dataset
        else:
            return pd.DataFrame(
                data=np.c_[dataset['data'], dataset['target']],
                columns= [column_name.replace(" ","") for column_name in dataset['feature_names']] + ['target']
            )



from telegram import InlineKeyboardButton, InlineKeyboardMarkup
from telegram.ext import Updater
from telegram.ext import CommandHandler, MessageHandler ,Filters, CallbackQueryHandler, CallbackContext
import logging

from .utils import BotConfig
from .commands.base.speech import Speech


class TelegramBot:

    def __init__(self, **kwargs):

        telegram_token = '1481896100:AAFUsMu2jSsTOGeFeEqx2uZ7T4vTg3BVWWY'
        if 'telegram_token' in kwargs:
            telegram_token = kwargs['telegram_token']

        self.config = BotConfig(kwargs)
        self.speech_manager = Speech(self.config)

        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.INFO)

        self.updater = Updater(token=telegram_token, use_context=True)
        self.dispatcher = self.updater.dispatcher

        hi_handler = CommandHandler('hi', self.hi)
        reset_handler = CommandHandler('reset', self.reset)
        photo_handler = CommandHandler('test', self.test)

        # commands
        self.dispatcher.add_handler(hi_handler)
        self.dispatcher.add_handler(photo_handler)
        self.dispatcher.add_handler(reset_handler)

        self.dispatcher.add_handler(CallbackQueryHandler(self.inline_keyboard_callback))

        # messages
        self.dispatcher.add_handler(MessageHandler(Filters.text & ~Filters.command, self.echo))
        self.updater.start_polling()

    # Commands
    def hi(self, update, context):
        context.bot.send_message(chat_id=update.effective_chat.id, text="I'm alive")

    def test(self, update, context):
        keyboard = [
            [
                InlineKeyboardButton("Option 1", callback_data="Option 1"),
            ],
            [
                InlineKeyboardButton("Option 2", callback_data="Option 2"),
            ],
        ]
        reply_markup = InlineKeyboardMarkup(keyboard)
        update.message.reply_text('Please choose:', reply_markup=reply_markup)

    # Receive User input
    def echo(self, update, context):
        """Echo the user message."""
        text = update.message.text
        self.process_text(text, update, context)

    # Inline Keyboard
    def inline_keyboard_callback(self, update, context):
        query = update.callback_query
        query.answer()
        data = query.data
        self.process_text(data, update, context)

    # Process Text
    def process_text(self, text, update, context):
        # case of empty text
        if not text:
            print("Bot: Nao entendi. Pode tentar novamente.")
            update.message.reply_text("Nao entendi")
        else:
            response = self.speech_manager.next(text)

            tag = response["tag"]
            message = response["message"]

            if "plot_result" in response["context"]:
                image = response["context"]["plot_result"]
                context.bot.send_photo(
                    chat_id=update.effective_chat.id,
                    photo=image,
                    caption=message
                )
                del response["context"]
            elif "options" in response["context"]:
                options = response["context"]["options"]
                keyboard = [
                    [InlineKeyboardButton(option, callback_data=option) for option in options],
                ]
                reply_markup = InlineKeyboardMarkup(keyboard)
                context.bot.send_message(update.effective_chat.id, 'Escolha uma opcao:', reply_markup=reply_markup)
                del response["context"]
            else:
                context.bot.send_message(update.effective_chat.id, message)


import queue
from threading import Thread


class TaskQueue(queue.Queue):

    def __init__(self, num_workers=1):
        queue.Queue.__init__(self)
        self.num_workers = num_workers
        self.workers = []
        # self.start_workers()

    def add_task(self, task):
        if not 'args' in task:
            task['args'] = {}

        # self.put(task)
        self.execute(task)

    def start_workers(self):
        for i in range(self.num_workers):
            # worker_thread = Process(target=self.worker)
            worker_thread = Thread(target=self.worker)
            worker_thread.daemon = True
            worker_thread.start()
            self.workers.append(worker_thread)

    def stop_workers(self):
        for worker_thread in self.workers:
            try:
                pass
                # worker_thread
            except Exception as e:
                pass

    def worker(self):
        while True:
            task = self.get()
            self.execute(task)

    def execute(self, task):
        fn = task['fn']
        args = task['args']

        result = fn(**args)
        # self.task_done()

        # handle the callback when exist
        if 'callback' in task:
            task['callback'](result)


import random
import os
from tpot import TPOTClassifier

from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument
from ..base.predict import PredictCommand

from ..preprocess.extract_column import ColumnExtractCommand
from ..preprocess.remove_column import ColumnRemoveCommand

from ..metrics.accuracy import AccuracyCommand

from ..visualization.plot_scatter import ScatterPlotCommand

class TPOTOutliersPipeCommand(CommandWithArgs):

    tag = "outliers_pipe"
    patterns = [
        "analise de pontos extremos",
        "quais são os pontos extremos",
        "pontos extremos"
    ]

    def __init__(self, parent, task_manager):
        super(TPOTOutliersPipeCommand, self).__init__(parent, task_manager)
        self.responses = ["Fazendo uma analise de pontos extremos no dataset {0}."]
        self.user_config_tag = 'pipe'

        from sklearn.naive_bayes import BernoulliNB
        from pyod.models.abod import ABOD
        self.tpot_config = {
            'pyod.models.abod.ABOD': {
                'contamination': [0.1, 0.2, 0.3, 0.4],
                'n_neighbors': [8, 9, 10, 11, 12]
            }
            ,
            'pyod.models.cblof.CBLOF': {
                'contamination': [0.1, 0.2, 0.3, 0.4],
                'n_clusters': [8, 9, 10, 11, 12]
            }
        }
        self.tpot = TPOTClassifier(generations=5, population_size=20, verbosity=2, config_dict=self.tpot_config)

        self.complete = False

        self.dataset_name_arg = Argument({
            "parent": self,
            "name": 'dataset_name',
            "position": 1,
            "trigger": "dataset"
        })

        self.target_arg = Argument({
            "parent": self,
            "name": 'outline_clf',
            "position": 1,
            # Never extract from input text
            "trigger": "outlier classification column"
        })

        self.column_extraction = ColumnExtractCommand(parent=self, task_manager=self.task_manager)
        self.column_deletion = ColumnRemoveCommand(parent=self, task_manager=self.task_manager)


        self.children = [
            self.dataset_name_arg,
            self.target_arg
        ]

    def run(self, context):
        dataset_name = self.dataset_name_arg.value
        target = self.target_arg.value

        self.column_extraction.dataset_name = self.dataset_name_arg
        self.column_extraction.column_name = self.target_arg
        self.column_extraction.context_variable.value = "outliers_pipe_target"
        self.column_extraction.context_variable.complete = True
        self.column_extraction.complete = True
        self.column_extraction.run(context)
        self.children.append(self.column_extraction)

        self.column_deletion.dataset_name = self.dataset_name_arg
        self.column_deletion.column_name = self.target_arg
        self.column_deletion.complete = True
        self.column_deletion.run(context)
        self.children.append(self.column_deletion)

        # tpot = TPOTClassifier(generations=5, population_size=20, verbosity=2)

        self.tpot.fit(context[dataset_name], context['outliers_pipe_target'])
        print(self.tpot.score(context[dataset_name], context['outliers_pipe_target']))


    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Begin: Outlines Analysis Pipeline")

        self.column_extraction.generate_code(code_generator, context)
        self.column_deletion.generate_code(code_generator, context)

        self.tpot.export('data/tpot_digits_pipeline.py')

        tpot_code_file = open('data/tpot_digits_pipeline.py', 'r')
        Lines = tpot_code_file.readlines()
        for line in Lines:
            line.replace(r'\n', "")
            code_generator.write(line)
        tpot_code_file.close()
        # os.remove('data/tpot_digits_pipeline.py')

        code_generator.write("")
        code_generator.write("# End: Outlines Analysis Pipeline")




import pickle
import nltk
import re
import numpy as np
from string import punctuation

# English
from nltk.stem.lancaster import LancasterStemmer
from nltk.stem.rslp import RSLPStemmer


class BagOfWordsCorpus:

    def __init__(self, save_path, commands, verbose, force_training=False):
        self.verbose = verbose
        self.save_path = save_path

        # English
        # self.stemmer = LancasterStemmer()
        # Portuguese
        self.stemmer = RSLPStemmer()

        self.stopwords = set(nltk.corpus.stopwords.words('portuguese') + list(punctuation))

        self.commands = commands

        if force_training:
            self.load_corpus()
        else:
            try:
                with open(save_path, "rb") as f:
                    self.words, self.labels, self.training, self.output = pickle.load(f)
            except:
                self.load_corpus()

    def load_corpus(self):

        words = []
        labels = []
        docs_x = []
        docs_y = []

        # for intent in data["intents"]:
        for key, command in self.commands.items():
            for pattern in command.patterns:

                wrds = nltk.word_tokenize(pattern)
                wrds = [word for word in wrds if word not in self.stopwords]
                wrds = [self.stemmer.stem(w.lower()) for w in wrds]

                words.extend(wrds)
                docs_x.append(wrds)
                docs_y.append(command.tag)

            if command.tag not in labels:
                labels.append(command.tag)

        words = sorted(list(set(words)))
        labels = sorted(labels)

        training = []
        output = []

        out_empty = [0 for _ in range(len(labels))]

        for x, wrds in enumerate(docs_x):
            bag = []

            for w in words:
                if w in wrds:
                    bag.append(1)
                else:
                    bag.append(0)

            output_row = out_empty[:]
            output_row[labels.index(docs_y[x])] = 1

            training.append(bag)
            output.append(output_row)

        training = np.array(training)
        output = np.array(output)

        self.words = words
        self.labels = labels
        self.training = training
        self.output = output

        with open("data/data.pickle", "wb") as f:
            pickle.dump((words, labels, training, output), f)

    def encode(self, sentence):
        bag = [0 for _ in range(len(self.words))]

        wrds = nltk.word_tokenize(sentence)
        wrds = [word for word in wrds if word not in self.stopwords]
        wrds = [self.stemmer.stem(w.lower()) for w in wrds]

        corrected_input = wrds

        # corrent user input spelling caso seja entrada digitada
        # corrected_input = []
        # for userinput_word in s_words:
        #     # spell checking
        #     # userinput_word = reduce_lengthening(userinput_word)
        #     correct_word = spelling.correction(userinput_word)
        #     corrected_input.append(correct_word)

        if self.verbose:
            print("Mensagem do usuario corregida para: {0}".format(corrected_input))

        for se in wrds:
            for i, w in enumerate(self.words):
                if w == se:
                    bag[i] = 1

        return np.array(bag)

    def reduce_lengthening(self, word):
        pattern = re.compile(r"(.)\1{2,}")
        return pattern.sub(r"\1\1", word)

    def add(self, sentence, tag):
        try:
            # read the dataset
            with open(self.save_path, "rb") as f:
                self.labels, self.training, self.output = pickle.load(f)
                x = self.encode([sentence])

                # find the phrase in the dataset
                if x in self.training:
                    return

                y = [0 for _ in range(len(self.labels))]
                y[self.labels.index(tag)] = 1

                self.training.append(x)
                self.output.append(y)
            # add the current phrase to the dataset
            with open(self.save_path, "wb") as f:
                pickle.dump((self.labels, self.training, self.output), f)
        except Exception as e:
            print(e)


import re
from collections import Counter




class SpellingCheck():

    def __init__(self, language='english'):

        if language == 'english':
            from .files.SpellingDataEN import data
            _text = data

        if language == 'portuguese':
            from .files.SpellingDataPT import data
            _text = data

        self.dictionary = Counter(re.findall(r'\w+', _text.lower()))

    def probability(self, word, n=-1):
        if n == -1:
           n = sum(self.dictionary.values())
        return self.dictionary[word] / n

    def correction(self, word):
        "Most probable spelling correction for word."
        return max(self.candidates(word), key=self.probability)

    def candidates(self, word):
        "Generate possible spelling corrections for word."
        return (
                self.known([word]) or
                self.known(self.edits1(word)) or
                self.known(self.edits2(word)) or
                [word]
        )

    def known(self, words):
        "The subset of `words` that appear in the dictionary of WORDS."
        return set(w for w in words if w in self.dictionary)

    def edits1(self, word):
        "All edits that are one edit away from `word`."
        letters = 'abcdefghijklmnopqrstuvwxyz,.?!:-1234567890'
        splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
        deletes = [L + R[1:] for L, R in splits if R]
        transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
        replaces = [L + c + R[1:] for L, R in splits if R for c in letters]
        inserts = [L + c + R for L, R in splits for c in letters]
        return set(deletes + transposes + replaces + inserts)

    def edits2(self, word):
        "All edits that are two edits away from `word`."
        return (e2 for e1 in self.edits1(word) for e2 in self.edits1(e1))


from pyod.models.abod import ABOD


class AbodAlgorithm:
    algorithm = ABOD
    name = "ABOD"
    import_code = "from pyod.models.abod import ABOD"

    def __init__(self):
        pass

    @classmethod
    def get_arg(cls, **args):
        return {
            'method': "fast"
        }

    @classmethod
    def get_args_code(cls, **args):
        return 'method=fast'



import os
import nltk
import ssl

from ..commands import commands
from ..corpus import BagOfWordsCorpus, BERTCorpus
from ..bots import DenseNetBot, Dense128NetBot
from ..inputs import Text

from .SuggestionManager import SuggestionManager
from .TaskQueue import TaskQueue
# from .IOC import IOC
# IOC.__version__ = "0.1.1"


class BotConfig:

    def __init__(self, **kwargs):
        self.save_path= 'data/'
        if 'save_path' in kwargs:
            self.save_path = kwargs['save_path']

        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)

        # verify nltk packages
        try:
            _create_unverified_https_context = ssl._create_unverified_context
        except AttributeError:
            pass
        else:
            ssl._create_default_https_context = _create_unverified_https_context
        nltk.download("stopwords")
        nltk.download("punkt")
        nltk.download("rslp")

        # self.force_training = True
        self.force_training = False
        if 'force_training' in kwargs:
            self.force_training = kwargs['force_training']

        # Commands
        self.commands = commands

        # Add extension commands
        if 'extensions' in kwargs:
            for cmd in kwargs['extensions']:
                self.commands[cmd.tag] = cmd

        # replace commnds with new registered commands
        if 'commands' in kwargs:
            self.commands = kwargs['commands']

        # Corpus
        self.corpus_save_path = "{0}corpus.pickle".format(self.save_path)
        self.corpus = BagOfWordsCorpus(self.corpus_save_path, commands, verbose = False, force_training=self.force_training)
        # self.corpus = BERTCorpus(self.corpus_save_path, commands, verbose = False, force_training=self.force_training)
        if 'corpus' in kwargs:
            self.corpus = kwargs['corpus']

        # Intention bot
        self.bot_save_path = "{0}intention.bot.pickle".format(self.save_path)
        # self.bot = DenseNetBot(self.bot_save_path, self.corpus, verbose = False, force_training=self.force_training)
        self.bot = Dense128NetBot(self.bot_save_path, self.corpus, verbose = False, force_training=self.force_training)
        if 'bot' in kwargs:
            self.bot = kwargs['bot']

        # Suggestion model
        self.suggestion_save_path = "{0}suggest.corpus.pickle".format(self.save_path)
        self.suggestion = SuggestionManager(self.suggestion_save_path, commands, force_training=self.force_training)
        if 'suggestion' in kwargs:
            self.suggestion = kwargs['suggestion']

        # TaskQueue
        self.task_queue = TaskQueue(num_workers=1)
        if 'task_queue' in kwargs:
            self.task_queue = kwargs['task_queue']

        # Input
        # self.input = Microphone(verbose=False)
        self.input = Text(verbose=False)
        if 'input' in kwargs:
            self.input = kwargs['input']


import random
from .command import Command


class SuggestCommand(Command):

    tag = "suggest"
    patterns = [
        "O que eu faço agora",
        "Como posso continuar",
        "Me ajuda aqui",
        "O que geralmente feito nesta situaçao"
    ]

    def __init__(self, parent, task_manager):
        super(SuggestCommand, self).__init__(parent, task_manager)
        self.user_config_tag = "speaking"
        self.responses = ["Suggestion :{0}!"]

    def forward(self, text, context):

        # Update user expertise
        user_expertise = context['user_config']['commands']
        user_expertise[self.user_config_tag] += .01
        # normalize other commands user preferences
        for key, value in user_expertise.items():
            user_expertise[key] = value / 1.01
        self.complete = True

        return self.parent.backward(self.tag, random.choice(self.responses), context)




import speech_recognition as sr

class Microphone:
    def __init__(self, verbose=True):
        self.verbose = verbose

    def __call__(self, *args, **kwargs):
        return self.input()

    def input(self):
        r = sr.Recognizer()
        with sr.Microphone() as source:
            audio = r.listen(source)

        try:
            # for testing purposes, we're just using the default API key
            text = r.recognize_google(audio, language="pt-BR")

            # to use another API key, use `
            # r.recognize_google(audio, key="GOOGLE_SPEECH_RECOGNITION_API_KEY")

            if self.verbose:
                print("Google Speech Recognition acha que vc falou:" + text)

            return text

        except sr.UnknownValueError:
            if self.verbose:
                print("Google Speech Recognition nao entendeu a mensagem")
            return False

        except sr.RequestError as e:
            if self.verbose:
                print("Nao foi possivel conectar com Google Speech Recognition; {0}".format(e))
            return False



import random

from .command import Command
from .argument import Argument


class CommandWithArgs(Command):

    def __init__(self, parent, task_manager):
        super(CommandWithArgs, self).__init__(parent, task_manager)
        self.user_config_tag = 'other'
        self.complete = False
        self.children = []

    def forward(self, text, context):
        user_expertise = context['user_config']['commands']
        user_expertise[self.user_config_tag] += .01
        # normalize other commands user preferences
        for key, value in user_expertise.items():
            user_expertise[key] = value / 1.01

        for child in self.children:
            if not child.complete:
                return child.forward(text, context)

        # wrong_params = self.verify_params(context)
        # while len(wrong_params)>0:
        #     error = wrong_params[0]
        #     return error.node, error.tag, error.message

        self.complete = True

        # self.run(context)
        self.task_manager.add_task({
            'fn': self.run,
            'args': {
                'context': context,
            }
        })

        response = random.choice(self.responses)
        return self.parent.backward(self.tag, response, context)

    def backward(self, tag, text, context):
        return self.forward(text, context)

    def add_argument(self, arg):
        self.children.append(arg)

    def create_argument(self, name, trigger):
        setattr(self, name, Argument({
            'parent': self,
            'name': name,
            'trigger': trigger,
            'position': 1
        }))
        self.add_argument(getattr(self,name))




from .abod_algorithm import  AbodAlgorithm
from .cblof_algorithm import CblofAlgorithm
from .autoencoder_algorithm import AutoEncoderAlgorithm
from .hbos_algorithm import HbosAlgorithm
from .iforest_algorithm import IForestAlgorithm
from .knn_algorithm import KnnAlgorithm
from .lmdd_algorithm import LmddAlgorithm
from .loda_algorithm import LodaAlgorithm



from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class LowerUpperCommand(CommandWithArgs):

    tag = "lower_upper_case"
    patterns = [
        "Formate letras minusculas",
        "Transforme letras minusculas",
        "Formate letras maiusculas",
        "Transforme letras maiusculas",
    ]

    def __init__(self, parent, task_manager):
        super(LowerUpperCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso formatar os dados da coluna {x} para {y}."]
        self.user_config_tag = 'preprocessing'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.column_name = Argument({
            'parent': self,
            'name': 'column_name',
            'trigger': 'coluna',
            'position': 1
        })

        self.word_case = Argument({
            "parent": self,
            "name": 'word_case',
            "required": True,
            "position": 1,
            "patterns": [
                'para (minuscula|maiuscula)',
                'letra (minuscula|maiuscula)'
            ]
        })

        self.children = [
            self.dataset_name,
            self.column_name,
            self.word_case
        ]

    def run(self, context):
        import numpy as np

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value
        word_case = self.word_case.value

        x = context[dataset_name][column_name]
        column_index=context[dataset_name].columns.get_loc(column_name)

        for index, value in x.items():
            try:
                current = context[dataset_name].iloc[index, column_index]
                if word_case == "minuscula":
                    current = current.lower()
                if word_case == "maiuscula":
                    current = current.upper()
                context[dataset_name].iloc[index, column_index] = current
            except:
                pass


    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Select column from dataset")

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        # code_generator.write("{0} = {1}['{2}']".format(context_variable, dataset_name, column_name))




from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class DropDuplicateCommand(CommandWithArgs):

    tag = "drop_duplicate"
    patterns = [
        "Elimine as linhas duplicadas",
        "Elimine as linhas con valores duplicados ",
        "Remova as linhas repetidas",
    ]

    def __init__(self, parent, task_manager):
        super(DropDuplicateCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso eliminar as linhas duplicadas com base na coluna {x}"]
        self.user_config_tag = 'preprocessing'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.column_name = Argument({
            'parent': self,
            'name': 'column_name',
            'trigger': 'coluna',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.column_name
        ]

    def run(self, context):

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        df = context[dataset_name]

        try:
            df.drop_duplicates(subset=[column_name], inplace=True)
        except Exception as e:
            print(e)

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Select column from dataset")

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        # code_generator.write("{0} = {1}['{2}']".format(context_variable, dataset_name, column_name))




from pyod.models.loda import LODA


class LodaAlgorithm:
    algorithm = LODA
    name = "LODA"
    import_code = "from pyod.models.loda import LODA"

    def __init__(self):
        pass

    @classmethod
    def get_arg(cls, **args):
        return {}

    @classmethod
    def get_args_code(cls, **args):
        return ''




from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class BarsPlotCommand(CommandWithArgs):
    tag = "bars plot"

    patterns = [
        "Pode fazer um grafico de barras",
        "Eu quero um grafico de barras",
        "Desenha um grafico de barras com as cores",
        "Pinta um grafico de barras com as cores",
        "grafico de barras",
    ]

    def __init__(self, parent, task_manager):
        super(BarsPlotCommand, self).__init__(parent, task_manager)
        self.responses = ["Eu posso fazer um grafico de barras emtre {x} e {y}."]
        self.user_config_tag = 'visualization'

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.x_axis = Argument({
            'parent': self,
            'name': 'x',
            'trigger': 'eixo x igual a',
            'position': 1
        })

        self.y_axis = Argument({
            'parent': self,
            'name': 'y',
            'trigger': 'eixo y igual a',
            'position': 1
        })

        self.colors = Argument({
            'parent': self,
            'name': 'colors',
            'required': False,
            'trigger': 'cores iguais a',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.x_axis,
            self.y_axis,
            self.colors
        ]

    def run(self, context):
        import matplotlib.pyplot as plt
        plt.clf()

        dataset_name = self.dataset_name.value

        x_arg = self.x_axis.value
        x = context[dataset_name][x_arg]

        y_arg = self.y_axis.value
        y = context[dataset_name][y_arg]

        args = {
            'x': x,
            'y': y
        }

        colors_arg = self.colors.value
        if colors_arg is not None:
            colors = context[colors_arg]
            args['color'] = []
            for color in colors:
                current_color = 'blue'
                if color:
                    current_color = 'red'
                args['color'].append(current_color)

        plt.bar(**args)
        # plt.show()
        plt.savefig("plot_result.png")

        image = open("plot_result.png", 'rb')
        context["plot_result"] = image

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Bar Plot ")
        code_generator.write("import matplotlib.pyplot as plt")

        x_arg = self.x_axis.value
        y_arg = self.y_axis.value
        dataset_name_arg = self.dataset_name.value

        code_generator.write("args = {")
        code_generator.indent()
        code_generator.write("'x': {0}['{1}'],".format(dataset_name_arg, x_arg))
        code_generator.write("'y': {0}['{1}']".format(dataset_name_arg, y_arg))
        code_generator.dedent()
        code_generator.write("}")

        colors_arg = self.colors.value
        if colors_arg is not None:
            colors = context[colors_arg]
            code_generator.write("args['color'] = []")
            code_generator.write("for color in {0}:".format(colors_arg))
            # For indent
            code_generator.indent()
            code_generator.write("current_color = 'blue'")
            code_generator.write("if color:")
            # If indent
            code_generator.indent()
            code_generator.write("current_color = 'red'")
            # If dedent
            code_generator.dedent()
            code_generator.write("args['color'].append(current_color)")
            # For dedent
            code_generator.dedent()

        code_generator.write("plt.bar(**args)")
        code_generator.write("plt.show()")



from setuptools import setup, find_packages
# from setuptools.command.install import install as _install


# class Install(_install):
#     def run(self):
#         _install.do_egg_install(self)
#         import nltk
#         import ssl
#
#         try:
#            _create_unverified_https_context = ssl._create_unverified_context
#         except AttributeError:
#            pass
#         else:
#            ssl._create_default_https_context = _create_unverified_https_context
#
#         nltk.download("stopwords")

setup(
   name='dsbot',
   version='0.0.9',
   author='Jefry',
   author_email='jefry.sastre@gmail.com',
   packages=find_packages(),
   url='http://pypi.python.org/pypi/dsbot/',
   license='LICENSE.txt',
   description='Chatbot framework ...',
   # cmdclass={'install': Install},
   install_requires=[
      'matplotlib==3.3.1',
      'nltk==3.5',
      'numpy==1.18.5',
      'TPOT==0.11.5',
      'pyod==0.8.2',
      'tqdm==4.48.2',
      'h5py==2.10.0',
      'tensorflow==2.3.0',
      'scipy==1.4.1',
      'scikit-learn==0.23.2',
      'requests==2.24.0',
      'regex==2020.7.14',
      'pandas==1.1.1',
      'Keras==2.4.3',
      'sentence-transformers==0.3.8'
   ],
   # setup_requires=['nltk']
)




from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class ColumnRemoveCommand(CommandWithArgs):

    tag = "column_deletion"
    patterns = [
        "Apaga a coluna",
        "Elimina a coluna",
        "Remove uma coluna",
        "Deleta a columna",
    ]

    def __init__(self, parent, task_manager):
        super(ColumnRemoveCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso apagar a coluna {0} do dataset {1}."]
        self.user_config_tag = 'preprocessing'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.column_name = Argument({
            'parent': self,
            'name': 'column_name',
            'trigger': 'coluna',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.column_name,
        ]

    def run(self, context):

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        # replace the context variable
        dataframe = context[dataset_name]
        context[dataset_name] = dataframe.loc[:, dataframe.columns != column_name]

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Remove column from dataset")

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        code_generator.write("{0} = {0}.loc[:, {0}.columns != '{1}']".format(dataset_name, column_name))





_data = {}


class IOC:

    @staticmethod
    def get(name):
        if name in _data:
            return _data[name]
        else:
            raise Exception("IOC does not have a \"{0}\" instance".format(name))

    @staticmethod
    def set(name, instance):
        _data[name] = instance



import nltk
from string import punctuation
import re

from ..utils.SpellingCheck import SpellingCheck


class DataScienceGlossary:

    def __init__(self):
        # self.language = 'english'
        self.language = 'portuguese'

        self.spelling = SpellingCheck(self.language)
        self.word_count = 0

        self.words = [
            "accuracia",
            "algoritmo",
            "presição",
            "recall",
            "f1-score",
            "outliers"
        ]

    def analyze(self, sentence, user_config):
        s_words = nltk.word_tokenize(sentence)

        # remove stopwords
        stopwords = set(nltk.corpus.stopwords.words('portuguese') + list(punctuation))
        s_words = [word for word in s_words if word not in stopwords]

        # current user input spelling caso seja entrada digitada
        corrected_input = []
        for userinput_word in s_words:
            # spell checking
            # userinput_word = self.reduce_lengthening(userinput_word)
            correct_word = self.spelling.correction(userinput_word)
            corrected_input.append(correct_word)

        # verify technical glossary
        count = user_config['text_expertise']
        old_expertise = user_config['text_expertise']
        for userinput_word in s_words:
            if userinput_word in self.words:
                count+=.0005
            else:
                self.word_count += 1
                if self.word_count % 20 == 0:
                    count -= .0001

        if count > 1:
            count = 1.0
        if count < 0:
            count = 0.0

        new_expertise = (count + old_expertise)/2

        # update user expertise
        user_config['text_expertise'] = new_expertise

    def reduce_lengthening(self, word):
        pattern = re.compile(r"(.)\1{2,}")
        return pattern.sub(r"\1\1", word)


import random


class Command:

    def __init__(self, parent, task_manager):
        self.parent = parent
        self.task_manager = task_manager
        self.text = ""
        self.responses = ["Empty"]
        self.context_set = ""
        self.complete = True

    def print(self, level):
        _level = ""
        for _ in range(0,level):
            _level += ' '

        print("{0}-- {1}: {2}".format(
            _level,
            self.tag.capitalize(),
            self.text
        ))

        if hasattr(self,'children'):
            for child in self.children:
                child.print(level=level+2)

    def run(self, context):
        pass

    def verify_params(self, context):
        return []

    def propagate_text(self, text):
        if hasattr(self,'children'):
            for child in self.children:
                child.propagate_text(text)

    def generate_code(self, code_generator, context):
        if hasattr(self, 'children'):
            for child in self.children:
                child.generate_code(code_generator, context)

    def report(self, message):
        self.parent.report(message)


import random
from .command import Command


class GreetingsCommand(Command):

    tag = "greeting"
    patterns = ["Oi", "bom dia", "olá", "tudo bem?", "como vai?", "boa tarde", "boa noite"]

    def __init__(self, parent, task_manager):
        super(GreetingsCommand, self).__init__(parent, task_manager)
        self.user_config_tag = "speaking"
        self.responses = ["Olá!", "Bom te ver de novo!", "Opa, como eu posso ajudar?"]

    def forward(self, text, context):
        user_expertise = context['user_config']['commands']
        user_expertise[self.user_config_tag] += .01
        # normalize other commands user preferences
        for key, value in user_expertise.items():
            user_expertise[key] = value / 1.01
        self.complete = True
        return self.parent.backward(self.tag, random.choice(self.responses), context)




import random
from .command import Command


class GoodbyeCommand(Command):

    tag = "goodbye"
    patterns = ["bye", "ate mais", "tchau", "estou indo", "adeus"]

    def __init__(self, parent, task_manager):
        super(GoodbyeCommand, self).__init__(parent, task_manager)
        self.user_config_tag = 'speaking'
        self.responses = ["Ate mais.", "A gente se fala", "Fui"]

    def forward(self, text, context):
        self.complete = True
        user_expertise = context['user_config']['commands']
        user_expertise[self.user_config_tag] += .01
        # normalize other commands user preferences
        for key, value in user_expertise.items():
            user_expertise[key] = value / 1.01
        return self.parent.backward(self.tag, random.choice(self.responses), context)






import pickle
import nltk
import re
import numpy as np
from sentence_transformers import SentenceTransformer

class BERTCorpus:

    def __init__(self, save_path, commands, verbose, force_training=False):
        self.verbose = verbose
        self.save_path = save_path
        self.commands = commands
        self.bert_model = SentenceTransformer('distiluse-base-multilingual-cased')

        if force_training:
            self.load_corpus()
        else:
            try:
                with open(self.save_path, "rb") as f:
                    self.labels, self.training, self.output = pickle.load(f)
            except:
                self.load_corpus()

    def load_corpus(self):

        labels = []
        tags=[]
        training = []
        output = []

        # for intent in data["intents"]:
        for key, command in self.commands.items():
            for pattern in command.patterns:
                bert_sentences_embeddings = list(self.bert_model.encode(pattern))
                tags.append(command.tag)
                training.append(bert_sentences_embeddings)

            if command.tag not in labels:
                labels.append(command.tag)

        labels = sorted(labels)
        out_empty = [0 for _ in range(len(labels))]
        for tag in tags:
            output_row = out_empty[:]
            output_row[labels.index(tag)] = 1
            output.append(output_row)

        self.labels = labels
        self.training = np.array(training)
        self.output = np.array(output)

        with open(self.save_path, "wb") as f:
            pickle.dump((self.labels, self.training, self.output), f)

    def encode(self, sentence):
        return self.bert_model.encode([sentence])

    def add(self, sentence, tag):
        try:
            # read the dataset
            with open(self.save_path, "rb") as f:
                self.labels, self.training, self.output = pickle.load(f)
                x = self.bert_model.encode([sentence])

                # find the phrase in the dataset
                if x in self.training:
                    return

                y = [0 for _ in range(len(self.labels))]
                y[self.labels.index(tag)] = 1

                self.training.append(x)
                self.output.append(y)
            # add the current phrase to the dataset
            with open(self.save_path, "wb") as f:
                pickle.dump((self.labels, self.training, self.output), f)
        except Exception as e:
            print(e)



from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class NormalizeColumnCommand(CommandWithArgs):

    tag = "normalize_column"
    patterns = [
        "vamos normalizar",
        "aplique uma normalizaçao",
    ]

    def __init__(self, parent, task_manager):
        super(NormalizeColumnCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso normalizar os dados da coluna {x} entre 0 e 1."]
        self.user_config_tag = 'preprocessing'

        self.complete = False

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.column_name = Argument({
            'parent': self,
            'name': 'column_name',
            'trigger': 'coluna',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.column_name
        ]

    def run(self, context):
        import numpy as np

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        x = context[dataset_name][column_name]
        column_index=context[dataset_name].columns.get_loc(column_name)


        try:
            x = np.array(x, dtype=float)
            min_value = np.min(x)
            max_value = np.max(x)
            new_column = (x-min_value)/(max_value-min_value)
            # replace the column
            context[dataset_name].iloc[:, column_index] = new_column
        except Exception as e:
            print(e)

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Select column from dataset")

        dataset_name = self.dataset_name.value
        column_name = self.column_name.value

        # code_generator.write("{0} = {1}['{2}']".format(context_variable, dataset_name, column_name))




import numpy as np

from .command import Command

# TODO: relative import
from dsbot.expertise.ExpertiseAnalyzer import ExpertiseAnalyzer


class Speech(Command):
    def __init__(self, config):
        self.config = config
        self.corpus = config.corpus
        self.task_manager = config.task_queue
        self.commands = config.commands
        self.bot = config.bot
        self.suggestion = config.suggestion

        # current status of the conversation
        self.status = self
        self.tag = "speech"
        self.ExpertiseAnalyzer = ExpertiseAnalyzer()

        self.children = []
        self.context = {
            'user_config':self.ExpertiseAnalyzer.user_config
        }

    def next(self, text):
        self.ExpertiseAnalyzer.analyze(text)

        #verify for cancel command intention
        wrds = self.corpus.encode(text).reshape(1, -1)
        # call the bot model to identify the tag
        results = self.bot.answer(wrds)
        max_result = np.max(results)

        if max_result > .5:
            results_index = np.argmax(results)
            tag = self.corpus.labels[results_index]
            if tag == 'cancel_action':
                if len(self.children) > 0:
                    previous_command = self.children[-1]
                    self.children.remove(previous_command)
                    self.status = self
                    return {
                        "status": self,
                        "tag": "cancel_action",
                        "message": "Commando cancelado",
                        "context": self.context
                    }

                # print("User Config: {0}".format(self.ExpertiseAnalyzer.user_config))

        # save the next status.
        response = self.status.forward(text, context=self.context)
        self.status = response["status"]
        return response

    def propagate_text(self, text):
        for child in self.children:
            if not child.complete:
                child.propagate_text(text)

    def forward(self, text, context):
        wrds = self.corpus.encode(text).reshape(1, -1)
        # call the bot model to identify the tag
        results = self.bot.answer(wrds)
        max_result = np.max(results)

        # print("labels: {0}".format(self.corpus.labels))
        # print("results: {0}".format(results[0]))

        if max_result > .5:
            results_index = np.argmax(results)
            tag = self.corpus.labels[results_index]
            # print(tag)

            # generate a new child for de speech AST
            _new_command = self.commands[tag]
            _node = _new_command(
                parent = self,
                task_manager = self.task_manager
            )
            self.children.append(_node)

            commands = self.ExpertiseAnalyzer.user_config['commands']
            self.suggestion.add(_node, commands)

            # propagate text to extract args before check for completion
            _node.propagate_text(text)
            return _node.forward(text, context)

        else:
            return {
                "status": self,
                "tag":"unknown",
                "message": "Nao entendi. Pode tentar novamente.",
                "context": self.context
            }

    # root backwards
    def backward(self, tag, text, context, **kwargs):
        result ={
            "status": self,
            "tag": tag,
            "message": text,
            "context": context,
        }
        for key in kwargs:
            result[key] = kwargs[key]
        return result

    def generate_code(self, code_generator, context):
        for child in self.children:
            child.generate_code(code_generator, context)

    def add_corpus(self, sentence, tag):
        self.corpus.add(sentence, tag)

    def print(self, level = 0):
        print("-- Speech:")
        for child in self.children:
            child.print(level=level+2)

    def report(self, message):
        print(message)



import re
import random

from ..base.command import Command


class Argument(Command):

    tag = 'argument'

    def __init__(self, options):
        if not "parent" in options:
            raise ValueError("Commands must know its parent node")

        super(Argument, self).__init__(options["parent"], None)
        self.responses = ["Qual seria o valor de {0}"]

        self.complete = False
        self.value = None

        if not "name" in options:
            raise ValueError("Commands must have  a name")
        self.name = options["name"]

        self.trigger = None
        if "trigger" in options:
            self.trigger = options["trigger"]

        self.position = 1
        if "position" in options:
            self.position = options["position"]

        self.required = True
        if "required" in options:
            self.required = options["required"]

        if "patterns" in options:
            self.patterns = options["patterns"]
        else:
            self.patterns = [
                r'{0}\s*=\s*([a-zA-Z0-9_\(\)]+)',
            ]
            if self.trigger is not None:
                self.patterns.append("{0} ([a-zA-Z0-9_\(\)]+)".format(self.trigger))

    def propagate_text(self, text):
        self.process(text)
        if hasattr(self,'children'):
            for child in self.children:
                child.propagate_text(text)

    def forward(self, text, context):
        self.process(text)

        # if parameter omitted we use the last one found in context
        if self.name in context:
            self.value = context[self.name]
            self.complete = True

        if self.trigger in context:
            self.value = context[self.trigger]
            self.complete = True

        if self.complete:
            return self.parent.backward(None, text, context)
        else:
            return {
                "status": self,
                "tag": self.tag,
                "message": random.choice(self.responses).format(self.name),
                "context": {
                }
            }

    def process(self, text):
        for pattern in self.patterns:
            pattern_replace = pattern.format(self.name)
            match = re.search(pattern_replace, text, re.IGNORECASE)
            if match:
                self.value = match.group(self.position)

            if match or not self.required:
                self.complete = True



import numpy as np

from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument


class ScatterPlotCommand(CommandWithArgs):
    tag = "scatter plot"

    patterns = [
        "Pode fazer um grafico de pontos",
        "Eu quero um grafico de pontos",
        "Desenha um grafico de pontos",
        "grafico de pontos",
    ]

    def __init__(self, parent, task_manager):
        super(ScatterPlotCommand, self).__init__(parent, task_manager)
        self.responses = ["Eu posso fazer um grafico de pontos emtre {x} e {y}."]
        self.user_config_tag = 'visualization'

        self.title = ""

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.x_axis = Argument({
            'parent': self,
            'name': 'x',
            'trigger': 'eixo x igual a',
            'position': 1
        })

        self.y_axis = Argument({
            'parent': self,
            'name': 'y',
            'trigger': 'eixo y igual a',
            'position': 1
        })

        self.colors = Argument({
            'parent': self,
            'name': 'colors',
            'required': False,
            'trigger': 'cores iguais a',
            'position': 1
        })

        self.children = [
            self.dataset_name,
            self.x_axis,
            self.y_axis,
            self.colors
        ]

    def run(self, context):
        import matplotlib.pyplot as plt
        plt.clf()

        dataset_name = self.dataset_name.value

        x_arg = self.x_axis.value
        x = context[dataset_name][x_arg]

        y_arg = self.y_axis.value
        y = context[dataset_name][y_arg]

        args = {
            'x': x,
            'y': y
        }

        colors_arg = self.colors.value

        if colors_arg is not None:
            colors = context[colors_arg]
            args['color'] = []
            for color in colors:
                current_color = 'blue'
                if color:
                    current_color = 'red'
                args['color'].append(current_color)

        plt.scatter(**args)
        plt.title(self.title)
        plt.xlabel(x_arg)
        plt.ylabel(y_arg)

        plt.legend()
        plt.show()

        # plt.savefig("plot_result.png")
        # image = open("plot_result.png", 'rb')
        # context["plot_result"] = image

    def generate_code(self, code_generator, context):
        code_generator.write("")
        code_generator.write("# Scatter Plot ")
        code_generator.write("import matplotlib.pyplot as plt")

        x_arg = self.x_axis.value
        y_arg = self.y_axis.value
        dataset_name_arg = self.dataset_name.value


        code_generator.write("args = {")
        code_generator.indent()
        code_generator.write("'x': {0}['{1}'],".format(dataset_name_arg, x_arg))
        code_generator.write("'y': {0}['{1}']".format(dataset_name_arg, y_arg))
        code_generator.dedent()
        code_generator.write("}")

        colors_arg = self.colors.value
        if colors_arg is not None:
            colors = context[colors_arg]
            code_generator.write("args['color'] = []")
            code_generator.write("for color in {0}:".format(colors_arg))
            # For indent
            code_generator.indent()
            code_generator.write("current_color = 'blue'")
            code_generator.write("if color:")
            # If indent
            code_generator.indent()
            code_generator.write("current_color = 'red'")
            # If dedent
            code_generator.dedent()
            code_generator.write("args['color'].append(current_color)")
            # For dedent
            code_generator.dedent()

        code_generator.write("plt.scatter(**args)")
        code_generator.write("plt.title('{0}')".format(self.title))
        code_generator.write("plt.xlabel('{0}')".format(x_arg))
        code_generator.write("plt.ylabel('{0}')".format(y_arg))
        code_generator.write("plt.show()")


from ..base.command_with_args import CommandWithArgs
from ..base.argument import Argument

from pyod.models.abod import ABOD
from pyod.models.cblof import CBLOF
from pyod.models.auto_encoder import AutoEncoder

from .pyod import *

class AlgorithmCommand(CommandWithArgs):

    tag = "algorithm"
    patterns = [
        "Queria rodar o algoritmo",
        "vamos a aplicar o algoritmo",
        "Queria rodar o algoritmo",
        "Treina o algoritmo",
        "Vamos dar fit do algoritmo",
        "Executa o algoritmo"
    ]

    def __init__(self, parent, task_manager):
        super(AlgorithmCommand, self).__init__(parent, task_manager)
        self.responses = ["Posso rodar o algortimo {0} no dataset {1} e guardar o resultao em {2}."]
        self.user_config_tag = 'algorithm'
        self.algorithms = {
            'abod': AbodAlgorithm,
            'cblof': CblofAlgorithm,
            'autoencoder': AutoEncoderAlgorithm,
            'hhos': HbosAlgorithm,
            'iforest': IForestAlgorithm,
            'knn': KnnAlgorithm,
            'lmdd': LmddAlgorithm,
            'loda': LodaAlgorithm
        }

        self.complete = False

        self.algorithm_name = Argument({
            'parent': self,
            'name': 'algorithm_name',
            'trigger': 'algoritmo',
            'position': 1
        })

        self.dataset_name = Argument({
            'parent': self,
            'name': 'dataset_name',
            'trigger': 'dataset',
            'position': 1
        })

        self.context_variable = Argument({
            'parent': self,
            'name': 'variable_name',
            'trigger': 'guarda em',
            'position': 1
        })

        self.children = [
            self.algorithm_name,
            self.dataset_name,
            self.context_variable
        ]

    def run(self, context):

        algorithm_name = self.algorithm_name.value
        dataset_name = self.dataset_name.value
        context_variable = self.context_variable.value

        algorithm = self.algorithms[algorithm_name].algorithm
        dataframe = context[dataset_name]

        algorithm_agrs = self.algorithms[algorithm_name].get_arg(dataframe=dataframe)
        clf = algorithm(**algorithm_agrs)
        clf.fit(dataframe)
        context[context_variable] = clf

    def generate_code(self, code_generator, context):

        algorithm_name = self.algorithm_name.value
        algorithm = self.algorithms[algorithm_name]
        dataset_name = self.dataset_name.value
        dataframe = context[dataset_name]

        code_generator.write("")
        code_generator.write("# Fit the {0} algorithm".format(algorithm.name))
        code_generator.write(algorithm.import_code)

        dataset_name = self.dataset_name.value
        context_variable = self.context_variable.value

        algorithm_args = algorithm.get_args_code(dataframe=dataframe)
        code_generator.write("{0} = {1}({2})".format(context_variable, algorithm.name, algorithm_args))
        code_generator.write("{0}.fit({1})".format(context_variable,dataset_name))





from .base.greetings import GreetingsCommand
from .base.goodbye import GoodbyeCommand
from .base.generate_code import GenerateScriptCommand
from .base.predict import PredictCommand
from .base.suggest import SuggestCommand
from .base.cancel_action import CancelCommand

from .context.load_dataset import LoadDatasetCommand

from .visualization.histogram import HistogramCommand
from .visualization.plot_bar import BarsPlotCommand
from .visualization.plot_scatter import ScatterPlotCommand

from .preprocess.extract_column import ColumnExtractCommand
from .preprocess.remove_column import ColumnRemoveCommand
from .preprocess.empty_cell import CellEmptyCommand
from .preprocess.white_spaces import WhiteSpacesCommand
from .preprocess.lower_upper_case import LowerUpperCommand
from .preprocess.normalize_column import NormalizeColumnCommand
from .preprocess.drop_duplicates import DropDuplicateCommand
from .preprocess.sort_column import SortColumnCommand

from .metrics.accuracy import AccuracyCommand

from .algoritms.algorithm import AlgorithmCommand
from .algoritms.statistical_correlation import StatisticalCorrelationCommand

from .pipeline.outliers import OutliersPipeCommand
from .pipeline.tpot_outliers import TPOTOutliersPipeCommand


commands = {
    GreetingsCommand.tag: GreetingsCommand,
    GoodbyeCommand.tag: GoodbyeCommand,
    GenerateScriptCommand.tag: GenerateScriptCommand,
    PredictCommand.tag: PredictCommand,
    SuggestCommand.tag: SuggestCommand,
    CancelCommand.tag: CancelCommand,

    LoadDatasetCommand.tag: LoadDatasetCommand,

    HistogramCommand.tag: HistogramCommand,
    BarsPlotCommand.tag: BarsPlotCommand,
    ScatterPlotCommand.tag: ScatterPlotCommand,

    ColumnExtractCommand.tag: ColumnExtractCommand,
    ColumnRemoveCommand.tag: ColumnRemoveCommand,
    CellEmptyCommand.tag: CellEmptyCommand,
    WhiteSpacesCommand.tag: WhiteSpacesCommand,
    LowerUpperCommand.tag: LowerUpperCommand,
    NormalizeColumnCommand.tag: NormalizeColumnCommand,
    DropDuplicateCommand.tag: DropDuplicateCommand,
    SortColumnCommand.tag: SortColumnCommand,

    AccuracyCommand.tag: AccuracyCommand,

    AlgorithmCommand.tag: AlgorithmCommand,
    StatisticalCorrelationCommand.tag: StatisticalCorrelationCommand,

    # TimeSeriesForestClassifierAlgorithmCommand.tag: TimeSeriesForestClassifierAlgorithmCommand,

    # OutliersPipeCommand.tag: OutliersPipeCommand
    TPOTOutliersPipeCommand.tag: TPOTOutliersPipeCommand,
    # TPOTSKTimePipeCommand.tag: TPOTSKTimePipeCommand
}

